///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2017 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                            //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////
// *************************************************************
// * ∇ - Temps logique ⇒ Temps discret - AHOSLA
// *************************************************************
// * Arbitrary high-order schemes for the linear advection eq.
// * S. Del Pino, H. Jourdren - doi:10.1016/j.crma.2006.01.013
// *************************************************************
with ℝ; // 1D

// Déclarations des options
ℕ Nx = 4;       // Nombre de mailles
ℝ a = 1.0;      // Coefficient d'advection
ℝ L = 1.0;      // Longueur totale de mailles
ℝ CFL = 1.0;    // Coefficient pour le calcul du δt
ℕ order = 2;    // Ordre de la méthode

// Déclaration des variables
ℝ t,δt,δx;      // Temps, pas-de-temps et pas-d'espace
ℝ x,c;          // Position des noeuds et des centres de mailles
ℝ u,F,ν;        // Solution, flux et coefficient (ν=a*δt/δx)

// Fonctions Maths - fact(N) & Cnp(n,p)
ℕ fact(ℕ n){
  ℕ f=1;
  for(ℕ i=2;i<=n;i+=1) f*=i;
  return f;
}
ℕ C(ℕ n, ℕ p){return fact(n)/(fact(p)*fact(n-p));}

// Calcul des constantes δx, δt & ν
δx = L/Nx;
δt = CFL*δx/a;
 ν = a*δt/δx;

// Calcul des positions des noeuds & des centres de mailles
∀ j∈cells, xⱼ=j*δx;
∀ j∈cells, ∀ r∈nodes(j) cⱼ=½*Σᵣxᵣ;

// Condition Initiale uⁿ⁼⁰ aux mailles
∀ j∈cells, uⱼⁿ⁼⁰=sin(2.0*π*cⱼ);
//∀ j∈cells uⱼⁿ⁼⁰=(cⱼ<0.5)?1.0:0.0;

// Calcul des flux Fⁿ⁽ᴺ⁼⁰⁾ et Fⁿ⁽ᴺ⁺¹⁾ aux noeuds
// On utilise la connectivité node_to_cell_left (n2cl) (1D)
∀ n∈ℕ, ∀ nodes Fⁿ⁽ᴺ⁼⁰⁾=0.0;
∀ n∈ℕ*, ∀ N∈[1..order], ∀ r∈nodes, j∈ℕ/j=n2cl(r), m∈ℕ/m=N/2
   Fᵣⁿ⁽ᴺ⁺¹⁾=Fᵣⁿ⁽ᴺ⁾-1.0/(fact(N+1))
   *∏_{i=m,i!=0}^{(N-1)/2}(ν+i)
     *Σ_{k=0}^{N}(-1.0)^(k+N+1)*C(N,k)*uⁿ_{j+m-k};
∀ n∈ℕ*, ∀ nodes Fⁿ=Fⁿ⁽ᴺ⁼ᵒʳᵈᵉʳ⁾;

// Calcul de uⁿ⁺¹ aux mailles
// Connectivités j+½ = node_right, j-½ = node_left (1D)
∀ n∈ℕ*, ∀ j∈cells uⱼⁿ⁺¹ = uⱼⁿ-ν*(Fⁿ_{j+½}-Fⁿ_{j-½});
